! This subroutine defines the values of the D array, which holds the inverted
! value of the distance to the (Nx+1, Ny+1, Nz+1) corner of the simulation box.
! This is used by the BCpotentials subroutine to compute the potential
! at the outer boundaries to be used as the boundary condition.

subroutine CalcDistance(D,Nx,Ny,Nz,dxyz,xflow,yflow,zflow)

  implicit none

! Parameters
  logical, intent(in) :: xflow, yflow, zflow
  integer, intent(in) :: Nx, Ny, Nz
  real*8, intent(in) :: dxyz(3)
  real*8, intent(out) :: D(2*Nx+1,2*Ny+1,2*Nz+1) ! 8 times the simulation box

! Local variables
  integer :: ii, jj, kk, pi, pj, pk, ni, nj, nk
  real*8 :: di, dj, dk, temp

  temp = 0.0d0
  D = 0.0d0

  do kk = 1, Nz+1
   do jj = 1, Ny+1
    do ii = 1, Nx+1
       pi = Nx+1+(ii-1)
       pj = Ny+1+(jj-1)
       pk = Nz+1+(kk-1)
       ni = Nx+1-(ii-1)
       nj = Ny+1-(jj-1)
       nk = Nz+1-(kk-1)
       di = dble(ii-1)
       dj = dble(jj-1)
       dk = dble(kk-1)

! Don't save zero distance
       if ( ii .eq. 1 .and. jj .eq. 1 .and. kk .eq. 1 ) then
          temp = 0.0d0
       else
          temp = dsqrt( (di*dxyz(1))**2.0d0 + (dj*dxyz(2))**2.0d0 + &
               (dk*dxyz(3))**2.0d0 )**(-1.0d0)
       end if

        ! If one of xflow, yflow and zflow is false, periodic boundary
        ! conditions are used in that dimension, and that makes for
        ! different numbers.
       if (xflow) then
          if (.not. yflow .and. .not. zflow) then
             temp = temp + dsqrt( (di*dxyz(1))**2.0d0 + &
                  ((Ny-dj)*dxyz(2))**2.0d0 + (dk*dxyz(3))**2.0d0 )**(-1.0d0)
             temp = temp + dsqrt( (di*dxyz(1))**2.0d0 + &
                  (dj*dxyz(2))**2.0d0 + ((Nz-dk)*dxyz(3))**2.0d0 )**(-1.0d0)
             temp = temp + dsqrt( (di*dxyz(1))**2.0d0 + &
                  ((Ny-dj)*dxyz(2))**2.0d0+((Nz-dk)*dxyz(3))**2.0d0 )**(-1.0d0)
          else if (.not. yflow) then
             temp = temp + dsqrt( (di*dxyz(1))**2.0d0 + &
                  ((Ny-dj)*dxyz(2))**2.0d0 + (dk*dxyz(3))**2.0d0 )**(-1.0d0)
          else if (.not. zflow) then
             temp = temp + dsqrt( (di*dxyz(1))**2.0d0 + (dj*dxyz(2))**2.0d0 &
                  + ((Nz-dk)*dxyz(3))**2.0d0 )**(-1.0d0)
          end if
       else if (yflow) then
          if (.not. zflow) then
             temp = temp + dsqrt( ((Nx-di)*dxyz(1))**2.0d0 + &
                  (dj*dxyz(2))**2.0d0 + (dk*dxyz(3))**2.0d0 )**(-1.0d0)
             temp = temp + dsqrt( (di*dxyz(1))**2.0d0      + &
                  (dj*dxyz(2))**2.0d0 + ((Nz-dk)*dxyz(3))**2.0d0 )**(-1.0d0)
             temp = temp + dsqrt( ((Nx-di)*dxyz(1))**2.0d0 + &
                  (dj*dxyz(2))**2.0d0 + ((Nz-dk)*dxyz(3))**2.0d0 )**(-1.0d0)
          else
             temp = temp + dsqrt( ((Nx-di)*dxyz(1))**2.0d0 + &
                  (dj*dxyz(2))**2.0d0 + (dk*dxyz(3))**2.0d0 )**(-1.0d0)
          end if
       else if (zflow) then
          temp = temp + dsqrt( ((Nx-di)*dxyz(1))**2.0d0 + (dj*dxyz(2))**2.0d0 &
               + (dk*dxyz(3))**2.0d0 )**(-1.0d0)
          temp = temp + dsqrt( (di*dxyz(1))**2.0d0      + &
               ((Ny-dj)*dxyz(2))**2.0d0 + (dk*dxyz(3))**2.0d0 )**(-1.0d0)
          temp = temp + dsqrt( ((Nx-di)*dxyz(1))**2.0d0 + &
               ((Ny-dj)*dxyz(2))**2.0d0 + (dk*dxyz(3))**2.0d0 )**(-1.0d0)
       else
          temp = temp + dsqrt( ((Nx-di)*dxyz(1))**2.0d0 + &
               (dj*dxyz(2))**2.0d0      + (dk*dxyz(3))**2.0d0 )**(-1.0d0)
          temp = temp + dsqrt( (di*dxyz(1))**2.0d0      + &
               ((Ny-dj)*dxyz(2))**2.0d0 + (dk*dxyz(3))**2.0d0 )**(-1.0d0)
          temp = temp + dsqrt( (di*dxyz(1))**2.0d0      + &
               (dj*dxyz(2))**2.0d0      + ((Nz-dk)*dxyz(3))**2.0d0 )**(-1.0d0)
          temp = temp + dsqrt( ((Nx-di)*dxyz(1))**2.0d0 + &
               ((Ny-dj)*dxyz(2))**2.0d0 + (dk*dxyz(3))**2.0d0 )**(-1.0d0)      
          temp = temp + dsqrt( (di*dxyz(1))**2.0d0      + &
               ((Ny-dj)*dxyz(2))**2.0d0 + ((Nz-dk)*dxyz(3))**2.0d0 )**(-1.0d0) 
          temp = temp + dsqrt( ((Nx-di)*dxyz(1))**2.0d0 + &
               (dj*dxyz(2))**2.0d0      + ((Nz-dk)*dxyz(3))**2.0d0 )**(-1.0d0)  
          temp = temp + dsqrt( ((Nx-di)*dxyz(1))**2.0d0 + &
               ((Ny-dj)*dxyz(2))**2.0d0 + ((Nz-dk)*dxyz(3))**2.0d0 )**(-1.0d0)
       end if

       ! The array actually holds 8 copies of all values.
       ! This is due to how indexing works in the BCpotentials
       ! subroutine, particularly for the case where periodic
       ! boundary conditions are used.
       D(pi,pj,pk) = 1.0d0*temp
       D(ni,pj,pk) = 1.0d0*temp
       D(pi,nj,pk) = 1.0d0*temp
       D(pi,pj,nk) = 1.0d0*temp
       D(ni,nj,pk) = 1.0d0*temp
       D(pi,nj,nk) = 1.0d0*temp
       D(ni,pj,nk) = 1.0d0*temp
       D(ni,nj,nk) = 1.0d0*temp

    end do
   end do
  end do


  return
end subroutine CalcDistance
