!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! dump2outp is a program that reads dump files saved by the
! picard program and saves regular particle output files
! the process number for which this is performed is specified
! as a command line argument, for example like this:
! $ dump2outp 123
! for process 123, which means the program will read file
! dumps/dump_particles.p00123.picard.dump
! 
! This means that the program reads only one and also only
! saves one file. If treating many files is desired, then 
! the program must be run multiple times.
!
! HG 2021-09-29
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

program dump2outp

  use hdf5
  use ifport
  use SpecificTypes
  implicit none


  logical :: writeParticles, writeFields, xflow, yflow, zflow, &
       updateEfield, startfromdumpfile
  integer :: Nx, Ny, Nz, iteration, iter_end, iter_start, &
       ii, jj, kk, ss, num_local, num_global, maxit, &
       dump_period_dump, dump_period_particles, dump_period_pprobes, &
       dump_period_fields, dump_start, write_period_screen, &
       Nretries, attempts, Bmodel, &
       Nprobes, fields_collected
  real*8  :: Lx, Ly, Lz, Lx_min, Lx_max, Ly_min, Ly_max, Lz_min, Lz_max, & 
       Lx_local, Ly_local, Lz_local, xmin, xmax, ymin, ymax, &
       zmin, zmax, dxyz(3), dt, maxerr, E0(4), &
       B0(4), B1, B2, r1, r2, &
       uE_local, uE_global, uP_local, uP_global, &
       uEB0_global, uE0_global, uB0_global, uP0_global, u0_global, &
       dV, ve(4), nucleusradius, flatradius, Galandradius, fadeoutradius

  type(particlearrays) particles
  type(particlespecies), allocatable :: species(:)
  type(EandPprobe), allocatable :: probe(:)

  integer :: myid, Nprocs, ierr

  integer status
  character inparg*5

  integer iprocs, jprocs, kprocs
  integer Nx_local, Ny_local, Nz_local
  integer Nspecies
  integer, parameter :: max_per_proc = 2**25

  real*8, allocatable :: U(:,:,:)

  logical ex
  logical(4) result
  character :: myidname*5, filename*100
  integer(hid_t) :: file_id       ! File identifier

  
! Number of retries for disk access 
  parameter(Nretries = 10)

! Initialize FORTRAN interface for hdf5
  call h5open_f(ierr)

! Get the general input parameters
  call GetGeneralInput(iter_end, startfromdumpfile, dump_period_dump, &
       dump_period_particles, dump_period_pprobes, &
       dump_period_fields, dump_start, write_period_screen, &
       nucleusradius, flatradius, Galandradius, fadeoutradius, &
       iprocs, jprocs, kprocs, Nx_local, Ny_local, Nz_local, Nspecies, &
       xmin, xmax, ymin, ymax, zmin, zmax, dt, &
       Bmodel, B0, B1, B2, r1, r2, Nprobes)

  ! Allocate the array for the species number of all particles
  allocate( particles%species(max_per_proc) )
  ! the array with positions and velocities
  allocate( particles%coordinates(6, max_per_proc) )
  ! and the species specific information.
  allocate( species(Nspecies) )
  ! and finally the potential matrix, which we don't really need, but
  ! allocating it allows us to use the LoadDump subroutine without
  ! any changes.
  allocate( U(Nx_local+2,Ny_local+2,Nz_local+2) )
  
  ! Now, read the input file again to get the species specific parameters.
  call GetSpecies(Nspecies, species)

  ! myid is the process number, given as an argument to this program
  call getarg(1,inparg)
  read(inparg,*,iostat=status) myid 

! Load the dump file
  attempts = 0
  call LoadDump(Nspecies, num_local, max_per_proc, &
       particles, myid, iteration, U, Nx_local, Ny_local, Nz_local, &
       attempts, Nretries)

! Make directory, check whether the output file exists and create it if not
  result = makedirqq('outp/particles')
  write(myidname,fmt='(I5.5)') myid
  filename = 'outp/particles/particles_'//'p'//myidname//'.h5'

  inquire(file=filename, exist=ex)
  if (.not. ex) then
     call h5fcreate_f(filename, h5f_acc_TRUNC_f, file_id, ierr)
     call h5fclose_f(file_id, ierr)
  end if
  
  ! Save the particle output file
  attempts = 0
  call DumpParticles( particles, num_local, iteration, myid, &
       attempts, Nretries )

! Close FORTRAN interface for hdf5
  call h5close_f(ierr)

end program dump2outp
