subroutine Gaussian(u, v, t, m)
    ! U is a random velocity from the Maxwellian with bulk velocity V, 
    ! and temperature T for a species of mass M. 
    implicit none
    real*8, intent(out) :: u(3)         ! [m/s]
    real*8, intent(in)  :: v(3), t, m   ! [m/s], [K], [kg]

    real*8 :: a, pw
    real*8, parameter :: kb   = 1.38064852d-23

    if (t .gt. 0.0d0 .and. m .gt. 0.0d0) then

      a = m/(2.0d0*kb*t)
      pw = dsqrt( a**(-1) )         ! approx. width of distribution

      call normal(u(1))
      call normal(u(2))
      call normal(u(3))
      u = u*pw/dsqrt(2.0d0)

      u = u + v

    else
      u = v
    end if

end subroutine Gaussian

!------------------------------------------------------------------------

subroutine normal(u)
    ! Gaussian distribution by Box-Muller. 
    ! We are wasting one random nmber, see 001219/boxmuller.c
    ! Also, it can be done without trig.  See NR. 
    implicit none
    real*8, intent(out) :: u
    real*8, parameter   :: tau  = 6.2831853071795864769252867665590d0  ! 2*pi
    real*8              :: r1, r2
    call random_number(r1)
    do
       call random_number(r2)
       if (r2 .gt. 0.0d0) then
         if (-2.0d0*dlog(r2) .ge. 0.0d0) then
           exit
         end if
       end if
    end do
    u = dsqrt(-2.0d0*dlog(r2))*dsin(tau*r1)
end subroutine normal

!------------------------------------------------------------------------

! This function gives us an ion density following Galand et al. (2016)
! with the exponential in the Haser model approximated by unity.

real*8 function Galand(nu, Q, r, nucleusradius, u)

  implicit none

  real*8, intent(in) :: nu, Q, r, nucleusradius, u

  real*8 pi

  parameter ( pi = 3.141592653589793d0 )

  Galand = nu * Q * (r-nucleusradius) / (4.0d0*pi*r**2.0d0*u**2.0d0)

end function Galand

!------------------------------------------------------------------------

! This function gives us a neutral density following Haser (1957)
! with the exponential approximated by unity. 

real*8 function Haser(nu, Q, r, nucleusradius, u)

  implicit none

  real*8, intent(in) :: nu, Q, r, nucleusradius, u

  real*8 pi

  parameter ( pi = 3.141592653589793d0 )

  Haser = Q / (4.0d0*pi*r**2.0d0*u)

  ! Here's the Haser model with the exponential included:
  !  Haser = ( Q / (4.0d0*pi*r**2.0d0*u) ) * exp(nu*(r-nucleusradius)/u)
  ! We don't use this in order to be consistent with the Galand model,
  ! which makes this approximation

end function Haser

!------------------------------------------------------------------------

! This function gives us a magnetic field profile that is high
! near the origin falls off over a transition region and reaches
! a lower value further out.

real*8 function Bfield(r, r1, r2, B1, B2)

  implicit none

  real*8, intent(in) :: r, r1, r2, B1, B2

!!$  if (r .le. r1) then
!!$     Bfield = B1
!!$  elseif (r .ge. r2) then
!!$     Bfield = B2
!!$  else
!!$     Bfield = 0.5d0*(B1+B2) - (B1-B2)*(r-0.5d0*(r1+r2))/(r2-r1);
!!$  end if

  ! Smoothed out version
  Bfield =  0.5d0*(B1+B2) - 0.5d0*(B1-B2)*tanh(2.0d0*(r-0.5d0*(r1+r2))/(r2-r1));

end function Bfield

!------------------------------------------------------------------------

